/*
 * COPYRIGHT:        See COPYRIGHT.TXT
 * PROJECT:          Ext2 File System Driver for WinNT/2K/XP
 * FILE:             dirctl.c
 * PROGRAMMER:       Matt Wu <mattwu@163.com>
 * HOMEPAGE:         http://ext2.yeah.net
 * UPDATE HISTORY: 
 */

/* INCLUDES *****************************************************************/

#include "ext2fs.h"

/* GLOBALS ***************************************************************/

extern PEXT2_GLOBAL Ext2Global;

/* DEFINITIONS *************************************************************/

#ifdef ALLOC_PRAGMA
#pragma alloc_text(PAGE, Ext2GetInfoLength)
#pragma alloc_text(PAGE, Ext2ProcessDirEntry)
#pragma alloc_text(PAGE, Ext2QueryDirectory)
#pragma alloc_text(PAGE, Ext2NotifyChangeDirectory)
#pragma alloc_text(PAGE, Ext2DirectoryControl)
#pragma alloc_text(PAGE, Ext2IsDirectoryEmpty)
#endif

ULONG
Ext2GetInfoLength(IN FILE_INFORMATION_CLASS  FileInformationClass)
{
    switch (FileInformationClass) {

    case FileDirectoryInformation:
        return FIELD_OFFSET(FILE_DIRECTORY_INFORMATION, FileName[0]);

    case FileFullDirectoryInformation:
        return FIELD_OFFSET(FILE_FULL_DIR_INFORMATION, FileName[0]);

    case FileBothDirectoryInformation:
        return FIELD_OFFSET(FILE_BOTH_DIR_INFORMATION, FileName[0]);

    case FileNamesInformation:
        return FIELD_OFFSET(FILE_NAMES_INFORMATION, FileName[0]);

    case FileIdFullDirectoryInformation:
        return FIELD_OFFSET(FILE_ID_FULL_DIR_INFORMATION, FileName[0]);

    case FileIdBothDirectoryInformation:
        return FIELD_OFFSET(FILE_ID_BOTH_DIR_INFORMATION, FileName[0]);

    default:
        break;
    }

    return 0;
}


NTSTATUS
Ext2ProcessDirEntry(
    IN PEXT2_IRP_CONTEXT    IrpContext,
    IN PEXT2_VCB            Vcb,
    IN PEXT2_FCB            Dcb,
    IN FILE_INFORMATION_CLASS  FileInformationClass,
    IN ULONG                in,
    IN PVOID                Buffer,
    IN ULONG                UsedLength,
    IN ULONG                Length,
    IN ULONG                FileIndex,
    IN PUNICODE_STRING      pName,
    OUT PULONG              EntrySize,
    IN BOOLEAN              Single
    )
{
    PEXT2_MCB   Mcb = NULL;
    PEXT2_MCB   Target = NULL;

    NTSTATUS    Status = STATUS_SUCCESS;
    EXT2_INODE inode;

    ULONG InfoLength = 0;
    ULONG NameLength = 0;
    ULONG dwBytes = 0;
    LONGLONG FileSize = 0;
    LONGLONG AllocationSize;

    *EntrySize = 0;
    NameLength = pName->Length;

    InfoLength = Ext2GetInfoLength(FileInformationClass);
    if (InfoLength == 0) {
        return STATUS_INVALID_INFO_CLASS;
    }

    if (InfoLength + NameLength > Length)  {
        DEBUG(DL_INF, ( "Ext2PricessDirEntry: Buffer is not enough.\n"));
        Status = STATUS_BUFFER_OVERFLOW;
        if (UsedLength || InfoLength > Length) {
            return Status;
        }
        RtlZeroMemory((PCHAR)Buffer + UsedLength, Length);
    } else {
        RtlZeroMemory((PCHAR)Buffer + UsedLength, InfoLength + NameLength);
    }

    DEBUG(DL_RES, ("Ext2ProcessDirEntry: %wZ\\%wZ\n", &Dcb->Mcb->FullName, pName));

    Mcb = Ext2SearchMcb(Vcb, Dcb->Mcb, pName);
    if (NULL == Mcb){

        if (!Ext2LoadInode(Vcb, in, &inode)) {
            DEBUG(DL_ERR, ( "Ext2PricessDirEntry: Loading inode %xh error.\n", in));
            DbgBreak();
            return STATUS_SUCCESS;
        }

        if (S_ISLNK(inode.i_mode)) {
            DEBUG(DL_RES, ("Ext2ProcessDirEntry: SymLink: %wZ\\%wZ\n",
                    &Dcb->Mcb->FullName, pName));
            Ext2LookupFile(IrpContext, Vcb, pName, Dcb->Mcb, &Mcb, 0);
        }
    }

    if (Mcb != NULL)  {

        if (IsMcbSymlink(Mcb)) {
            Target = Mcb->Target;
            ASSERT(Target);
            if (IsMcbDirectory(Target)) {
                FileSize = 0;
            } else {
                FileSize = Target->FileSize.QuadPart;
            }
        } else {
            if (IsMcbDirectory(Mcb)) {
                FileSize = 0;
            } else {
                FileSize = Mcb->FileSize.QuadPart;
            }
        }

    } else {
        if (S_ISDIR(inode.i_mode)) {
            FileSize = 0;
        } else {
            FileSize  = (LONGLONG) inode.i_size;
            if (S_ISREG(inode.i_mode)) {
                FileSize |= ((LONGLONG)(inode.i_size_high) << 32);
            }
        }
    }

    AllocationSize = CEILING_ALIGNED(ULONGLONG, FileSize, BLOCK_SIZE);

    switch(FileInformationClass) {

    case FileDirectoryInformation:

    {
        PFILE_DIRECTORY_INFORMATION     FDI;

        FDI = (PFILE_DIRECTORY_INFORMATION) ((PUCHAR)Buffer + UsedLength);
        if (!Single) {
            FDI->NextEntryOffset = InfoLength + NameLength;
        }

        FDI->FileIndex = FileIndex;

        if (Mcb) {

            FDI->CreationTime = Mcb->CreationTime;
            FDI->LastAccessTime = Mcb->LastAccessTime;
            FDI->LastWriteTime = Mcb->LastWriteTime;
            FDI->ChangeTime = Mcb->ChangeTime;
            FDI->FileAttributes = Mcb->FileAttr;

        } else {

            FDI->CreationTime = Ext2NtTime(inode.i_ctime);
            FDI->LastAccessTime = Ext2NtTime(inode.i_atime);
            FDI->LastWriteTime = Ext2NtTime(inode.i_mtime);
            FDI->ChangeTime = Ext2NtTime(inode.i_mtime);

            FDI->FileAttributes = FILE_ATTRIBUTE_NORMAL;

            if (Ext2IsReadOnly(inode.i_mode)) {
                SetFlag(FDI->FileAttributes, FILE_ATTRIBUTE_READONLY);
            }

            if (S_ISDIR(inode.i_mode))
                FDI->FileAttributes |= FILE_ATTRIBUTE_DIRECTORY;
        }

        FDI->EndOfFile.QuadPart = FileSize;
        FDI->AllocationSize.QuadPart = AllocationSize;

        FDI->FileNameLength = NameLength;
        if (InfoLength + NameLength > Length) {
            NameLength = Length - InfoLength;
        }
        RtlCopyMemory(FDI->FileName, pName->Buffer, NameLength);

        *EntrySize = InfoLength + NameLength; 
        break;
    }
       
    case FileFullDirectoryInformation:

    {
        PFILE_FULL_DIR_INFORMATION      FFI;

        FFI = (PFILE_FULL_DIR_INFORMATION) ((PUCHAR)Buffer + UsedLength);
        if (!Single) {
            FFI->NextEntryOffset = InfoLength + NameLength;
        }

        FFI->FileIndex = FileIndex;

        if (Mcb) {

            FFI->CreationTime = Mcb->CreationTime;
            FFI->LastAccessTime = Mcb->LastAccessTime;
            FFI->LastWriteTime = Mcb->LastWriteTime;
            FFI->ChangeTime = Mcb->ChangeTime;
            FFI->FileAttributes = Mcb->FileAttr;

        } else {

            FFI->CreationTime = Ext2NtTime(inode.i_ctime);
            FFI->LastAccessTime = Ext2NtTime(inode.i_atime);
            FFI->LastWriteTime = Ext2NtTime(inode.i_mtime);
            FFI->ChangeTime = Ext2NtTime(inode.i_mtime);

            FFI->FileAttributes = FILE_ATTRIBUTE_NORMAL;

            if (Ext2IsReadOnly(inode.i_mode)) {
                SetFlag(FFI->FileAttributes, FILE_ATTRIBUTE_READONLY);
            }

            if (S_ISDIR(inode.i_mode))
                FFI->FileAttributes |= FILE_ATTRIBUTE_DIRECTORY;
        }

        FFI->FileNameLength = NameLength;
        if (InfoLength + NameLength > Length) {
            NameLength = Length - InfoLength;
        }
        RtlCopyMemory(FFI->FileName, pName->Buffer, NameLength);

        *EntrySize = InfoLength + NameLength; 
        break;
    }
        
    case FileBothDirectoryInformation:

    {
        PFILE_BOTH_DIR_INFORMATION      FBI;

        FBI = (PFILE_BOTH_DIR_INFORMATION) ((PUCHAR)Buffer + UsedLength);
        if (!Single) {
            FBI->NextEntryOffset = InfoLength + NameLength;
        }

        FBI->FileIndex = FileIndex;
        FBI->EndOfFile.QuadPart = FileSize;
        FBI->AllocationSize.QuadPart = AllocationSize;

        if (Mcb) {

            FBI->CreationTime = Mcb->CreationTime;
            FBI->LastAccessTime = Mcb->LastAccessTime;
            FBI->LastWriteTime = Mcb->LastWriteTime;
            FBI->ChangeTime = Mcb->ChangeTime;
            FBI->FileAttributes = Mcb->FileAttr;

        } else {

            FBI->CreationTime = Ext2NtTime(inode.i_ctime);
            FBI->LastAccessTime = Ext2NtTime(inode.i_atime);
            FBI->LastWriteTime = Ext2NtTime(inode.i_mtime);
            FBI->ChangeTime = Ext2NtTime(inode.i_mtime);

            FBI->FileAttributes = FILE_ATTRIBUTE_NORMAL;

            if (Ext2IsReadOnly(inode.i_mode)) {
                SetFlag(FBI->FileAttributes, FILE_ATTRIBUTE_READONLY);
            }

            if (S_ISDIR(inode.i_mode)) {
                FBI->FileAttributes |= FILE_ATTRIBUTE_DIRECTORY;
            }
        }

        if (NameLength <= 24) {
            RtlCopyMemory(FBI->ShortName, pName->Buffer, NameLength);
            FBI->ShortNameLength = (UCHAR)NameLength;
        }

        FBI->FileNameLength = NameLength;
        if (InfoLength + NameLength > Length) {
            NameLength = Length - InfoLength;
        }
        RtlCopyMemory(FBI->FileName, pName->Buffer, NameLength);

        *EntrySize = InfoLength + NameLength; 
        break;
    }
        
    case FileNamesInformation:

    {
        PFILE_NAMES_INFORMATION         FNI;

        FNI = (PFILE_NAMES_INFORMATION) ((PUCHAR)Buffer + UsedLength);
        if (!Single) {
            FNI->NextEntryOffset = InfoLength + NameLength;
        }

        FNI->FileNameLength = NameLength;
        if (InfoLength + NameLength > Length) {
            NameLength = Length - InfoLength;
        }
        RtlCopyMemory(FNI->FileName, pName->Buffer, NameLength);

        *EntrySize = InfoLength + NameLength; 
        break;
    }

    case FileIdFullDirectoryInformation:

    {
        PFILE_ID_FULL_DIR_INFORMATION   FIF;

        FIF = (PFILE_ID_FULL_DIR_INFORMATION)((PUCHAR)Buffer + UsedLength);
        if (!Single) {
            FIF->NextEntryOffset = InfoLength + NameLength;
        }

        FIF->FileIndex = FileIndex;

        if (Mcb) {

            FIF->CreationTime = Mcb->CreationTime;
            FIF->LastAccessTime = Mcb->LastAccessTime;
            FIF->LastWriteTime = Mcb->LastWriteTime;
            FIF->ChangeTime = Mcb->ChangeTime;
            FIF->FileAttributes = Mcb->FileAttr;

        } else {

            FIF->CreationTime = Ext2NtTime(inode.i_ctime);
            FIF->LastAccessTime = Ext2NtTime(inode.i_atime);
            FIF->LastWriteTime = Ext2NtTime(inode.i_mtime);
            FIF->ChangeTime = Ext2NtTime(inode.i_mtime);

            FIF->FileAttributes = FILE_ATTRIBUTE_NORMAL;

            if (Ext2IsReadOnly(inode.i_mode)) {
                SetFlag(FIF->FileAttributes, FILE_ATTRIBUTE_READONLY);
            }

            if (S_ISDIR(inode.i_mode))
                FIF->FileAttributes |= FILE_ATTRIBUTE_DIRECTORY;
        }
        FIF->FileId.QuadPart = (LONGLONG)in;
        FIF->FileNameLength = NameLength;
        if (InfoLength + NameLength > Length) {
            NameLength = Length - InfoLength;
        }
        RtlCopyMemory(FIF->FileName, pName->Buffer, NameLength);

        *EntrySize = InfoLength + NameLength; 
        break;
    }

    case FileIdBothDirectoryInformation:

    {
        PFILE_ID_BOTH_DIR_INFORMATION   FIB;

        FIB = (PFILE_ID_BOTH_DIR_INFORMATION)((PUCHAR)Buffer + UsedLength);
        if (!Single) {
            FIB->NextEntryOffset = InfoLength + NameLength;
        }

        FIB->FileIndex = FileIndex;
        FIB->EndOfFile.QuadPart = FileSize;
        FIB->AllocationSize.QuadPart = AllocationSize;

        if (Mcb) {

            FIB->CreationTime = Mcb->CreationTime;
            FIB->LastAccessTime = Mcb->LastAccessTime;
            FIB->LastWriteTime = Mcb->LastWriteTime;
            FIB->ChangeTime = Mcb->ChangeTime;
            FIB->FileAttributes = Mcb->FileAttr;

        } else {

            FIB->CreationTime = Ext2NtTime(inode.i_ctime);
            FIB->LastAccessTime = Ext2NtTime(inode.i_atime);
            FIB->LastWriteTime = Ext2NtTime(inode.i_mtime);
            FIB->ChangeTime = Ext2NtTime(inode.i_mtime);

            FIB->FileAttributes = FILE_ATTRIBUTE_NORMAL;

            if (Ext2IsReadOnly(inode.i_mode)) {
                SetFlag(FIB->FileAttributes, FILE_ATTRIBUTE_READONLY);
            }

            if (S_ISDIR(inode.i_mode)) {
                FIB->FileAttributes |= FILE_ATTRIBUTE_DIRECTORY;
            }
        }
        FIB->FileId.QuadPart = (LONGLONG)in;

        if (NameLength <= 24) {
            RtlCopyMemory(FIB->ShortName, pName->Buffer, NameLength);
            FIB->ShortNameLength = (UCHAR)NameLength;
        }

        FIB->FileNameLength = NameLength;
        if (InfoLength + NameLength > Length) {
            NameLength = Length - InfoLength;
        }
        RtlCopyMemory(FIB->FileName, pName->Buffer, NameLength);

        *EntrySize = InfoLength + NameLength; 
        break;
    }

    default:
        Status = STATUS_INVALID_INFO_CLASS;
        break;
    }

    if (Mcb) {
        Ext2DerefMcb(Mcb);
    }

    return Status;
}


BOOLEAN
Ext2IsWearingCloak(
    IN  PEXT2_VCB       Vcb,
    IN  POEM_STRING     OemName
    )
{
    size_t  PatLen = 0;

    /* we could not filter the files: "." and ".." */
    if (OemName->Length >= 1 && OemName->Buffer[0] == '.') {

        if ( OemName->Length == 2 && OemName->Buffer[1] == '.') {
            return FALSE;
        } else if (OemName->Length == 1) {
            return FALSE;
        }
    }

    /* checking name prefix */
    if (Vcb->bHidingPrefix) {
        PatLen = strlen(&Vcb->sHidingPrefix[0]);
        if (PatLen <= OemName->Length) {
            if ( _strnicmp( OemName->Buffer, 
                            Vcb->sHidingPrefix,
                            PatLen ) == 0) {
                return TRUE;
            }
        }
    }

    /* checking name suffix */
    if (Vcb->bHidingSuffix) {
        PatLen = strlen(&Vcb->sHidingSuffix[0]);
        if (PatLen <= OemName->Length) {
            if ( _strnicmp(&OemName->Buffer[OemName->Length - PatLen], 
                            Vcb->sHidingSuffix, PatLen ) == 0) {
                return TRUE;
            }
        }
    }

    return FALSE;
}

NTSTATUS
Ext2QueryDirectory (IN PEXT2_IRP_CONTEXT IrpContext)
{
    PDEVICE_OBJECT          DeviceObject;
    NTSTATUS                Status = STATUS_UNSUCCESSFUL;
    PEXT2_VCB               Vcb;
    PFILE_OBJECT            FileObject;
    PEXT2_FCB               Fcb;
    PEXT2_CCB               Ccb;
    PIRP                    Irp;
    PIO_STACK_LOCATION      IoStackLocation;

    ULONG                   Length;
    ULONG                   FileIndex;
    PUNICODE_STRING         FileName;
    PUCHAR                  Buffer;

    BOOLEAN                 RestartScan;
    BOOLEAN                 ReturnSingleEntry;
    BOOLEAN                 IndexSpecified;
    BOOLEAN                 FirstQuery;
    BOOLEAN                 FcbResourceAcquired = FALSE;

    USHORT                  InodeFileNameLength;
    FILE_INFORMATION_CLASS  FileInformationClass;
    ULONG                   UsedLength = 0;

    OEM_STRING              OemName;
    UNICODE_STRING          InodeFileName;
    PEXT2_DIR_ENTRY2        pDir = NULL;

    ULONG                   ByteOffset;
    ULONG                   RecLen = 0;
    ULONG                   PrevEntry = 0;
    ULONG                   EntrySize = 0;

    InodeFileName.Buffer = NULL;
    
    __try {

        ASSERT(IrpContext);
        
        ASSERT((IrpContext->Identifier.Type == EXT2ICX) &&
            (IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));
        
        DeviceObject = IrpContext->DeviceObject;
        
        //
        // This request is not allowed on the main device object
        //
        if (IsExt2FsDevice(DeviceObject)) {
            Status = STATUS_INVALID_DEVICE_REQUEST;
            __leave;
        }
        
        Vcb = (PEXT2_VCB) DeviceObject->DeviceExtension;
        
        ASSERT(Vcb != NULL);
        
        ASSERT((Vcb->Identifier.Type == EXT2VCB) &&
            (Vcb->Identifier.Size == sizeof(EXT2_VCB)));

        ASSERT(IsMounted(Vcb));

        FileObject = IrpContext->FileObject;
        
        Fcb = (PEXT2_FCB) FileObject->FsContext;
        
        if (Fcb == NULL) {
            Status = STATUS_INVALID_PARAMETER;
            __leave;
        }
        
        //
        // This request is not allowed on volumes
        //
        if (Fcb->Identifier.Type == EXT2VCB) {
            Status = STATUS_INVALID_PARAMETER;
            __leave;
        }
        
        ASSERT((Fcb->Identifier.Type == EXT2FCB) &&
            (Fcb->Identifier.Size == sizeof(EXT2_FCB)));
        
        if (!IsDirectory(Fcb)) {
            Status = STATUS_NOT_A_DIRECTORY;
            __leave;
        }
        
        Ccb = (PEXT2_CCB) FileObject->FsContext2;
        
        ASSERT(Ccb);
        
        ASSERT((Ccb->Identifier.Type == EXT2CCB) &&
            (Ccb->Identifier.Size == sizeof(EXT2_CCB)));
        
        Irp = IrpContext->Irp;
        
        IoStackLocation = IoGetCurrentIrpStackLocation(Irp);
        
#ifndef _GNU_NTIFS_
        
        FileInformationClass =
            IoStackLocation->Parameters.QueryDirectory.FileInformationClass;
        
        Length = IoStackLocation->Parameters.QueryDirectory.Length;
        
        FileName = (PUNICODE_STRING)IoStackLocation->Parameters.QueryDirectory.FileName;
        
        FileIndex = IoStackLocation->Parameters.QueryDirectory.FileIndex;
        
#else // _GNU_NTIFS_
        
        FileInformationClass = ((PEXTENDED_IO_STACK_LOCATION)
            IoStackLocation)->Parameters.QueryDirectory.FileInformationClass;
        
        Length = ((PEXTENDED_IO_STACK_LOCATION)
            IoStackLocation)->Parameters.QueryDirectory.Length;
        
        FileName = ((PEXTENDED_IO_STACK_LOCATION)
            IoStackLocation)->Parameters.QueryDirectory.FileName;
        
        FileIndex = ((PEXTENDED_IO_STACK_LOCATION)
            IoStackLocation)->Parameters.QueryDirectory.FileIndex;
        
#endif // _GNU_NTIFS_

        RestartScan = FlagOn(((PEXTENDED_IO_STACK_LOCATION)
            IoStackLocation)->Flags, SL_RESTART_SCAN);
        ReturnSingleEntry = FlagOn(((PEXTENDED_IO_STACK_LOCATION)
            IoStackLocation)->Flags, SL_RETURN_SINGLE_ENTRY);
        IndexSpecified = FlagOn(((PEXTENDED_IO_STACK_LOCATION)
            IoStackLocation)->Flags, SL_INDEX_SPECIFIED);

        Buffer = Ext2GetUserBuffer(Irp);

        if (Buffer == NULL) {
            DbgBreak();
            Status = STATUS_INVALID_USER_BUFFER;
            __leave;
        }
        
        if (!IsFlagOn(IrpContext->Flags, IRP_CONTEXT_FLAG_WAIT)) {
            Status = STATUS_PENDING;
            __leave;
        }

        if (!ExAcquireResourceSharedLite(
                 &Fcb->MainResource,
                 IsFlagOn(IrpContext->Flags, IRP_CONTEXT_FLAG_WAIT) )) {
            Status = STATUS_PENDING;
            __leave;
        }
        FcbResourceAcquired = TRUE;
        
        if (FileName != NULL) {

            if (Ccb->DirectorySearchPattern.Buffer != NULL) {

                FirstQuery = FALSE;

            } else {

                FirstQuery = TRUE;

                Ccb->DirectorySearchPattern.Length =
                    Ccb->DirectorySearchPattern.MaximumLength =
                    FileName->Length;
                
                Ccb->DirectorySearchPattern.Buffer =
                    ExAllocatePoolWithTag(PagedPool, FileName->Length,
                                          EXT2_DIRSP_MAGIC);
                
                if (Ccb->DirectorySearchPattern.Buffer == NULL) {
                    DEBUG(DL_ERR, ( "Ex2QueryDirectory: failed to allocate SerarchPattern.\n")); 
                    Status = STATUS_INSUFFICIENT_RESOURCES;
                    __leave;
                }

                INC_MEM_COUNT( PS_DIR_PATTERN,
                               Ccb->DirectorySearchPattern.Buffer,
                               Ccb->DirectorySearchPattern.MaximumLength);

                Status = RtlUpcaseUnicodeString(
                    &(Ccb->DirectorySearchPattern),
                    FileName,
                    FALSE);

                if (!NT_SUCCESS(Status)) {
                    __leave;
                }
            }

        } else if (Ccb->DirectorySearchPattern.Buffer != NULL) {

            FirstQuery = FALSE;
            FileName = &Ccb->DirectorySearchPattern;

        } else {

            FirstQuery = TRUE;
            
            Ccb->DirectorySearchPattern.Length =
                Ccb->DirectorySearchPattern.MaximumLength = 2;
            
            Ccb->DirectorySearchPattern.Buffer =
                ExAllocatePoolWithTag(PagedPool, 4, EXT2_DIRSP_MAGIC);
            
            if (Ccb->DirectorySearchPattern.Buffer == NULL) {
                DEBUG(DL_ERR, ( "Ex2QueryDirectory: failed to allocate SerarchPattern (1st).\n")); 
                Status = STATUS_INSUFFICIENT_RESOURCES;
                __leave;
            }
            
            INC_MEM_COUNT( PS_DIR_PATTERN,
                           Ccb->DirectorySearchPattern.Buffer,
                           Ccb->DirectorySearchPattern.MaximumLength);

            RtlZeroMemory(Ccb->DirectorySearchPattern.Buffer, 4);
            RtlCopyMemory(
                Ccb->DirectorySearchPattern.Buffer,
                L"*\0", 2);
        }
        
        if (!IndexSpecified) {
            if (RestartScan || FirstQuery) {
                FileIndex = 0;
            } else {
                FileIndex = Ccb->CurrentByteOffset;
            }
        }
       
        RtlZeroMemory(Buffer, Length);

        if (Fcb->Inode->i_size <= FileIndex) {
            Status = STATUS_NO_MORE_FILES;
            __leave;
        }
        
        pDir = ExAllocatePoolWithTag(
                    PagedPool,
                    sizeof(EXT2_DIR_ENTRY2),
                    EXT2_DENTRY_MAGIC
                );

        if (!pDir) {
            DEBUG(DL_ERR, ( "Ex2QueryDirectory: failed to allocate pDir.\n")); 
            Status = STATUS_INSUFFICIENT_RESOURCES;
            __leave;
        }

        INC_MEM_COUNT(PS_DIR_ENTRY, pDir, sizeof(EXT2_DIR_ENTRY2));
        
        ByteOffset = FileIndex;
        PrevEntry = 0;
        
        while ((UsedLength < Length) && (ByteOffset < Fcb->Inode->i_size)) {

            RtlZeroMemory(pDir, sizeof(EXT2_DIR_ENTRY2));

            Status = Ext2ReadInode(
                        IrpContext,
                        Vcb,
                        Fcb->Mcb,
                        (ULONGLONG)ByteOffset,
                        (PVOID)pDir,
                        sizeof(EXT2_DIR_ENTRY2),
                        FALSE,
                        &EntrySize);

            if (!NT_SUCCESS(Status)) {
                __leave;
            }

            if (pDir->rec_len == 0) {
                RecLen = BLOCK_SIZE - (ByteOffset & (BLOCK_SIZE - 1));
            } else {
                RecLen = pDir->rec_len;
            }

            if (!pDir->inode || pDir->inode >= INODES_COUNT) {
                goto ProcessNextEntry;
            }

            OemName.Buffer = pDir->name;
            OemName.Length = (pDir->name_len & 0xff);
            OemName.MaximumLength = OemName.Length;

            if (Ext2IsWearingCloak(Vcb, &OemName)) {
                goto ProcessNextEntry;
            }

            InodeFileNameLength = (USHORT)
                Ext2OEMToUnicodeSize(Vcb, &OemName);
         
            if (InodeFileNameLength <= 0) {
                break;
            }

            if ( InodeFileName.Buffer != NULL &&
                 InodeFileName.MaximumLength > InodeFileNameLength) {
                /* reuse buffer */
            } else {
                /* free and re-allocate it */
                if (InodeFileName.Buffer) {
                    DEC_MEM_COUNT(PS_INODE_NAME,
                                  InodeFileName.Buffer,
                                  InodeFileName.MaximumLength);
                    ExFreePoolWithTag(InodeFileName.Buffer, EXT2_INAME_MAGIC);
                    InodeFileName.Buffer = NULL;
                }
                InodeFileName.MaximumLength = InodeFileNameLength + 2;
                InodeFileName.Buffer = ExAllocatePoolWithTag(
                        PagedPool,
                        InodeFileName.MaximumLength,
                        EXT2_INAME_MAGIC
                    );

                if (!InodeFileName.Buffer) {
                    DEBUG(DL_ERR, ( "Ex2QueryDirectory: failed to "
                                          "allocate InodeFileName.\n")); 
                    Status = STATUS_INSUFFICIENT_RESOURCES;
                    __leave;
                }
                INC_MEM_COUNT(PS_INODE_NAME, InodeFileName.Buffer,
                              InodeFileName.MaximumLength);
            }

            InodeFileName.Length = 0;
            RtlZeroMemory(InodeFileName.Buffer, InodeFileName.MaximumLength);
            
            Status = Ext2OEMToUnicode( Vcb,
                                       &InodeFileName,
                                       &OemName );

            if (!NT_SUCCESS(Status)) {
                DEBUG(DL_ERR, ( "Ex2QueryDirectory: Ext2OEMtoUnicode fialed with %xh.\n", Status)); 
                __leave;
            }

            if (FsRtlDoesNameContainWildCards(
                &(Ccb->DirectorySearchPattern)) ?
                    FsRtlIsNameInExpression(
                        &(Ccb->DirectorySearchPattern),
                        &InodeFileName,
                        TRUE,
                        NULL) :
                !RtlCompareUnicodeString(
                    &(Ccb->DirectorySearchPattern),
                    &InodeFileName,
                    TRUE)           ) {

                Status = Ext2ProcessDirEntry(
                                IrpContext,
                                Vcb,
                                Fcb,
                                FileInformationClass,
                                pDir->inode,
                                Buffer,
                                UsedLength, 
                                Length - UsedLength,
                                ByteOffset,
                                &InodeFileName,
                                &EntrySize,
                                ReturnSingleEntry
                                );

                if (NT_SUCCESS(Status)) {
                    if (EntrySize > 0) {
                        PrevEntry = UsedLength;
                        UsedLength += EntrySize;
                    } else {
                        DbgBreak();
                    }
                } else {
                    if (Status != STATUS_BUFFER_OVERFLOW) {
                        __leave;
                    }
                    break;
                }
            }

ProcessNextEntry:
            
            ByteOffset += RecLen;
            Ccb->CurrentByteOffset = ByteOffset;

            if (UsedLength && ReturnSingleEntry) {
                Status = STATUS_SUCCESS;
                __leave;
            }
        }

        FileIndex = ByteOffset;
        ((PULONG)((PUCHAR)Buffer + PrevEntry))[0] = 0;

        if (!UsedLength) {
            if (NT_SUCCESS(Status)) {
                if (FirstQuery) {
                    Status = STATUS_NO_SUCH_FILE;
                } else {
                    Status = STATUS_NO_MORE_FILES;
                }
            } else {
                ASSERT(Status == STATUS_BUFFER_OVERFLOW);
                UsedLength = EntrySize;
            }
        } else {
            Status = STATUS_SUCCESS;
        }

    } __finally {
    
        if (FcbResourceAcquired) {
            ExReleaseResourceLite(&Fcb->MainResource);
        }
        
        if (pDir != NULL) {
            ExFreePoolWithTag(pDir, EXT2_DENTRY_MAGIC);
            DEC_MEM_COUNT(PS_DIR_ENTRY, pDir, sizeof(EXT2_DIR_ENTRY2)); 
        }
        
        if (InodeFileName.Buffer != NULL) {
            DEC_MEM_COUNT(PS_INODE_NAME, InodeFileName.Buffer,
                          InodeFileName.MaximumLength );
            ExFreePoolWithTag(InodeFileName.Buffer, EXT2_INAME_MAGIC);
        }
        
        if (!IrpContext->ExceptionInProgress) {

            if ( Status == STATUS_PENDING || 
                 Status == STATUS_CANT_WAIT) {

                Status = Ext2LockUserBuffer(
                    IrpContext->Irp,
                    Length,
                    IoWriteAccess );
                
                if (NT_SUCCESS(Status)) {
                    Status = Ext2QueueRequest(IrpContext);
                } else {
                    Ext2CompleteIrpContext(IrpContext, Status);
                }
            } else {
                IrpContext->Irp->IoStatus.Information = UsedLength;
                Ext2CompleteIrpContext(IrpContext, Status);
            }
        }
    }
    
    return Status;
}

NTSTATUS
Ext2NotifyChangeDirectory (
    IN PEXT2_IRP_CONTEXT IrpContext
    )
{
    PDEVICE_OBJECT      DeviceObject;
    BOOLEAN             CompleteRequest = TRUE;
    NTSTATUS            Status = STATUS_UNSUCCESSFUL;
    PEXT2_VCB           Vcb;
    PFILE_OBJECT        FileObject;
    PEXT2_FCB           Fcb;
    PIRP                Irp;
    PIO_STACK_LOCATION  IrpSp;
    ULONG               CompletionFilter;
    BOOLEAN             WatchTree;

    BOOLEAN             bFcbAcquired = FALSE;

    __try {

        ASSERT(IrpContext);

        ASSERT((IrpContext->Identifier.Type == EXT2ICX) &&
               (IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));

        //
        //  Always set the wait flag in the Irp context for the original request.
        //

        SetFlag( IrpContext->Flags, IRP_CONTEXT_FLAG_WAIT );

        DeviceObject = IrpContext->DeviceObject;

        if (IsExt2FsDevice(DeviceObject)) {
            Status = STATUS_INVALID_DEVICE_REQUEST;
            __leave;
        }

        Vcb = (PEXT2_VCB) DeviceObject->DeviceExtension;

        ASSERT(Vcb != NULL);

        ASSERT((Vcb->Identifier.Type == EXT2VCB) &&
               (Vcb->Identifier.Size == sizeof(EXT2_VCB)));

        ASSERT(IsMounted(Vcb));

        FileObject = IrpContext->FileObject;

        Fcb = (PEXT2_FCB) FileObject->FsContext;

        ASSERT(Fcb);

        if (Fcb->Identifier.Type == EXT2VCB) {
            DbgBreak();
            Status = STATUS_INVALID_PARAMETER;
            __leave;
        }

        ASSERT((Fcb->Identifier.Type == EXT2FCB) &&
               (Fcb->Identifier.Size == sizeof(EXT2_FCB)));

        if (!IsDirectory(Fcb)) {
            DbgBreak();
            Status = STATUS_NOT_A_DIRECTORY;
            __leave;
        }

        if (ExAcquireResourceExclusiveLite(
                &Fcb->MainResource,
                TRUE ))  {
            bFcbAcquired = TRUE;
        } else {
            Status = STATUS_PENDING;
            __leave;
        }

        Irp = IrpContext->Irp;

        IrpSp = IoGetCurrentIrpStackLocation(Irp);

#ifndef _GNU_NTIFS_

        CompletionFilter =
            IrpSp->Parameters.NotifyDirectory.CompletionFilter;

#else // _GNU_NTIFS_

        CompletionFilter = ((PEXTENDED_IO_STACK_LOCATION)
            IrpSp)->Parameters.NotifyDirectory.CompletionFilter;

#endif // _GNU_NTIFS_

        WatchTree = IsFlagOn(IrpSp->Flags, SL_WATCH_TREE);

        if (FlagOn(Fcb->Flags, FCB_DELETE_PENDING)) {
            Status = STATUS_DELETE_PENDING;
            __leave;
        }

        FsRtlNotifyFullChangeDirectory( Vcb->NotifySync,
                                        &Vcb->NotifyList,
                                        FileObject->FsContext2,
                                        (PSTRING)(&Fcb->Mcb->FullName),
                                        WatchTree,
                                        FALSE,
                                        CompletionFilter,
                                        Irp,
                                        NULL,
                                        NULL );

        CompleteRequest = FALSE;

        Status = STATUS_PENDING;

/*
    Currently the driver is read-only but here is an example on how to use the
    FsRtl-functions to report a change:

    ANSI_STRING TestString;
    USHORT      FileNamePartLength;

    RtlInitAnsiString(&TestString, "\\ntifs.h");

    FileNamePartLength = 7;

    FsRtlNotifyReportChange(
        Vcb->NotifySync,            // PNOTIFY_SYNC NotifySync
        &Vcb->NotifyList,           // PLIST_ENTRY  NotifyList
        &TestString,                // PSTRING      FullTargetName
        &FileNamePartLength,        // PUSHORT      FileNamePartLength
        FILE_NOTIFY_CHANGE_NAME     // ULONG        FilterMatch
        );

    or

    ANSI_STRING TestString;

    RtlInitAnsiString(&TestString, "\\ntifs.h");

    FsRtlNotifyFullReportChange(
        Vcb->NotifySync,            // PNOTIFY_SYNC NotifySync
        &Vcb->NotifyList,           // PLIST_ENTRY  NotifyList
        &TestString,                // PSTRING      FullTargetName
        1,                          // USHORT       TargetNameOffset
        NULL,                       // PSTRING      StreamName OPTIONAL
        NULL,                       // PSTRING      NormalizedParentName OPTIONAL
        FILE_NOTIFY_CHANGE_NAME,    // ULONG        FilterMatch
        0,                          // ULONG        Action
        NULL                        // PVOID        TargetContext
        );
*/

    } __finally {

        if (bFcbAcquired) {
            ExReleaseResourceLite(&Fcb->MainResource);
        }

        if (!IrpContext->ExceptionInProgress) {
            if (CompleteRequest) {
                if (Status == STATUS_PENDING) {
                    Ext2QueueRequest(IrpContext);
                } else {
                    Ext2CompleteIrpContext(IrpContext, Status);
                }
            } else {
                IrpContext->Irp = NULL;
                Ext2CompleteIrpContext(IrpContext, Status);
            }
        }
    }

    return Status;
}

VOID
Ext2NotifyReportChange (
    IN PEXT2_IRP_CONTEXT IrpContext,
    IN PEXT2_VCB         Vcb,
    IN PEXT2_MCB         Mcb,
    IN ULONG             Filter,
    IN ULONG             Action   )
{
    USHORT          Offset;

    Offset = (USHORT) ( Mcb->FullName.Length - 
                        Mcb->ShortName.Length);

    FsRtlNotifyFullReportChange( Vcb->NotifySync,
                                 &(Vcb->NotifyList),
                                 (PSTRING) (&Mcb->FullName),
                                 (USHORT) Offset,
                                 (PSTRING)NULL,
                                 (PSTRING) NULL,
                                 (ULONG) Filter,
                                 (ULONG) Action,
                                 (PVOID) NULL );

    // ASSERT(KeGetCurrentIrql() < DISPATCH_LEVEL);
}


NTSTATUS
Ext2DirectoryControl (IN PEXT2_IRP_CONTEXT IrpContext)
{
    NTSTATUS Status;
    
    ASSERT(IrpContext);
    
    ASSERT((IrpContext->Identifier.Type == EXT2ICX) &&
        (IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));
    
    switch (IrpContext->MinorFunction) {

    case IRP_MN_QUERY_DIRECTORY:
        Status = Ext2QueryDirectory(IrpContext);
        break;

    case IRP_MN_NOTIFY_CHANGE_DIRECTORY:
        Status = Ext2NotifyChangeDirectory(IrpContext);
        break;
        
    default:
        Status = STATUS_INVALID_DEVICE_REQUEST;
        Ext2CompleteIrpContext(IrpContext, Status);
    }
    
    return Status;
}


BOOLEAN
Ext2IsDirectoryEmpty (
    PEXT2_IRP_CONTEXT   IrpContext,
    PEXT2_VCB           Vcb,
    PEXT2_MCB           Mcb
    )
{
    NTSTATUS                Status = STATUS_UNSUCCESSFUL;

    PEXT2_DIR_ENTRY2        pTarget = NULL;

    ULONG                   ByteOffset = 0;
    ULONG                   dwRet;

    BOOLEAN                 bEmpty = FALSE;

    if (!IsMcbDirectory(Mcb) || IsMcbSymlink(Mcb)) {
        return TRUE;
    }

    __try {

        pTarget = (PEXT2_DIR_ENTRY2)
                    ExAllocatePoolWithTag(
                        PagedPool,
                        EXT2_DIR_REC_LEN(EXT2_NAME_LEN),
                        EXT2_DENTRY_MAGIC
                    );
        if (!pTarget) {
            DEBUG(DL_ERR, ( "Ex2isDirectoryEmpty: failed to allocate pTarget.\n"));
            Status = STATUS_INSUFFICIENT_RESOURCES;
            __leave;
        }

        ByteOffset = 0;

        while (ByteOffset < Mcb->Inode.i_size) {

            RtlZeroMemory(pTarget, EXT2_DIR_REC_LEN(EXT2_NAME_LEN));

            Status = Ext2ReadInode(
                        IrpContext,
                        Vcb,
                        Mcb,
                        (ULONGLONG)ByteOffset,
                        (PVOID)pTarget,
                        EXT2_DIR_REC_LEN(EXT2_NAME_LEN),
                        FALSE,
                        &dwRet);

            if (!NT_SUCCESS(Status)) {
                DEBUG(DL_ERR, ( "Ext2IsDirectoryEmpty: Reading Directory Content error.\n"));
                __leave;
            }

            if (pTarget->inode) {
                if (pTarget->name_len == 1 && pTarget->name[0] == '.') {
                } else if (pTarget->name_len == 2 && pTarget->name[0] == '.' && 
                         pTarget->name[1] == '.') {
                } else {
                    break;
                }
            }

            ByteOffset += pTarget->rec_len;
        }

        bEmpty = (ByteOffset >= Mcb->Inode.i_size);

    } __finally {

        if (pTarget != NULL) {
            ExFreePoolWithTag(pTarget, EXT2_DENTRY_MAGIC);
        }
    }
    
    return bEmpty;
}
